%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Comparison of the experimental and the theoretical relative noise errors
%
%Takahiro Sasaki, Jan 2020, sasa0027@umn.edu
% Scientific Computation Program, Electrical and Computer Engineering Dept.
% The University of Minnesota Twin Cities
%
%This code is a part of our JOSA A journal paper:
% Sasaki, T., & Leger, J., "Light field reconstruction
%  from scattered light using plenoptic data," JOSA A (2020)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%{
***How to use***

    1)Choose a scattering material by changing "sampleID".
        0: Brushed metal, 1: White paper #1, 2: White paper #2
    2)Execute.

***Data format and some details***

    -Angular domain range is [29deg,61deg] for both theta_in and theta_out.
     Number of angular data points Nth is 64, or angular increment is 0.5deg.

    -pCOS.csv (Scattered signal from a periodic pattern displayed on the LCD):
        -This includes 500 trials of measurement.
        -Nth X 500 2-D array.
        -Row direction: 500 trials.
        -Column direction: Outgoing angle or theta_out.
        Note: Camera noise standard deviation is computed
                from a part of this data.

    -BRDFCOS.csv (BRDF x COS(theta_in) of a scatter)
        -Nth x Nth 2-D array.
        -Row direction: Outgoing angle, or theta_out.
        -Column direction: Incoming angle, or theta_in.
        -This includes COS(theta_in) term.

    -Spec.csv (Specular component of BRDF):
        -This needs to be computed from BRDF w/o COS term.
        -See section 5.A for details.
        -Nth 1-D array.

    -Diff.csv (Diffuse component of BRDF):
        -This needs to be computed from BRDF w/o COS term.
        -The diffuse is expressed as (Diff)(x)(Diff) where (x) represents
         outer-product, and data for (Diff) is the Diff.csv file.
        -See section 5.A for details.
        -Nth 1-D array.

***Notes***

    -The above BRDF-related files are slightly different
            from those for LightFieldRecon.m tool.

    -The provided data and results may be slightly different
        from those reported in our JOSA A journal paper.

***For further details, see the following codes and values***
%}

clear;

%%%%%%%%%%%%%%% Choose scattering material %%%%%%%%%%%%%
%0: Brushed metal, 1: White paper #1, 2: White paper #2
sampleID = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%% Common Settings %%%
%Number of data points in theta (angular) direction
Nth = 64;

%Number of trials to compute statistics.
numTrials = 500;

%%% Scatter dependent settings %%%
if(sampleID == 0)%For Brushed Metal scatter
   
    disp('*** Brushed metal scatter ***');
    
    %Number of measured signal files (pCOS.csv) to be considered.
    numPCOS = 4;

    %Number of COS periods displayed on LCD for each measured signal.
    pCOSAll = zeros(numPCOS,1);
    pCOSAll(1) = 10;%10 periods of COS pattern displayed on LCD
    pCOSAll(2) = 15;
    pCOSAll(3) = 20;
    pCOSAll(4) = 25;
    
    %File name of BRDF x COS. (Note: This includes COS(theta_in) term.)
    fNameBRDFCOS = './BrushedMetal/DeltaP/BRDFCOSMetalB.csv';
    
    %File name of measured signal files (pCOS.csv) including many trials.
    %Note: Signal mean and noise standard deviation are computed from one of these.
    fNameLmeasAll = './BrushedMetal/DeltaP/pCOSMetal';
    
    %File name of specular component of BRDF.
    fNameSpec = './BrushedMetal/DeltaP/SpecMetalB.csv';

    %File name of diffuse component of BRDF.
    %Note: This is separable expression. The diffuse is expressed as
    % (Diff)(X)(Diff) where (X) represents outer product, and (Diff) is this file:
    fNameDiff = './BrushedMetal/DeltaP/DiffMetalB.csv';     
    
elseif(sampleID == 1)%For White paper #1 scatter
    
    disp('*** White paper #1 scatter ***');

    %Number of measured signal files (pCOS.csv) to be considered.
    numPCOS = 8;

    %Number of COS periods displayed on LCD for each measured signal.
    pCOSAll = zeros(numPCOS,1);
    pCOSAll(1) = 2;%2 periods of COS pattern displayed on LCD
    pCOSAll(2) = 4;
    pCOSAll(3) = 6;
    pCOSAll(4) = 8;
    pCOSAll(5) = 10;
    pCOSAll(6) = 12;
    pCOSAll(7) = 13;
    pCOSAll(8) = 14;
    
    %File name of BRDF x COS. (Note: This includes COS(theta_in) term.)
    fNameBRDFCOS = './WhitePaper1/DeltaP/BRDFCOSWPaper1B.csv';
    
    %File name of measured signal files (pCOS.csv) including many trials.
    %Note: Signal mean and noise standard deviation are computed from one of these.
    fNameLmeasAll = './WhitePaper1/DeltaP/pCOSWPaper1_';
    
    %File name of specular component of BRDF.
    fNameSpec = './WhitePaper1/DeltaP/SpecWPaper1B.csv';

    %File name of diffuse component of BRDF.
    %Note: This is separable expression. The diffuse is expressed as
    % (Diff)(X)(Diff) where (X) represents outer product, and (Diff) is this file:
    fNameDiff = './WhitePaper1/DeltaP/DiffWPaper1B.csv';     
    
elseif(sampleID == 2)%For White paper #2 scatter
    
    disp('*** White paper #2 scatter ***');

    %Number of measured signal files (pCOS.csv) to be considered.
    numPCOS = 5;

    %Number of COS periods displayed on LCD for each measured signal.
    pCOSAll = zeros(numPCOS,1); 
    pCOSAll(1) = 2;%2 periods of COS pattern displayed on LCD
    pCOSAll(2) = 4;
    pCOSAll(3) = 6;
    pCOSAll(4) = 8;
    pCOSAll(5) = 10;
    
    %File name of BRDF x COS. (Note: This includes COS(theta_in) term.)
    fNameBRDFCOS = './WhitePaper2/DeltaP/BRDFCOSWPaper2B.csv';
    
    %File name of measured signal files (pCOS.csv) including many trials.
    %Note: Signal mean and noise standard deviation are computed from one of these.
    fNameLmeasAll = './WhitePaper2/DeltaP/pCOSWPaper2_';
    
    %File name of specular component of BRDF.
    fNameSpec = './WhitePaper2/DeltaP/SpecWPaper2B.csv';

    %File name of diffuse component of BRDF.
    %Note: This is separable expression. The diffuse is expressed as
    % (Diff)(X)(Diff) where (X) represents outer product, and (Diff) is this file:
    fNameDiff = './WhitePaper2/DeltaP/DiffWPaper2B.csv';     
end

%%% Other settings %%%

%Offset to compute standard deviation accurately. 
offset = 10;

%This is to select an experimental result to show as an example.
p2Show = 2;
%%% End of Settings %%%



%%% Compute experimental DeltaP from eq. (50) %%%

%Arrays to store results.
ratio = zeros(1, numTrials);
DeltaPExpe = zeros(2,numPCOS);

%Load data and prepare system matrix
Ker = csvread(fNameBRDFCOS);
FTKer = fliplr(fftshift(fft2(Ker)));
%Note: Center of FTKer is (Nth/2+1,Nth/2)

for p=1:1:numPCOS
    
    %Number of periods in current data.
    pCOS = pCOSAll(p);
    
    %Load Lmeas data including many trials.
    LmeasAll = csvread(strcat(fNameLmeasAll,num2str(pCOS),'.csv'));

    %Prepare truncated FTKer matrix.
    FTKerTrunc = FTKer((Nth/2+1)-pCOS:(Nth/2+1)+pCOS, (Nth/2)-pCOS:(Nth/2)+pCOS);

    for nSTD=1:1:numTrials

        %Pick one of trials of Lmeas.
        Lmeas = transpose(LmeasAll(nSTD,:));
       
        %Prepare truncated 1D-FT of Lmeas.
        FTLM = fftshift(fft(Lmeas));
        FTLMTrunc = FTLM((Nth/2+1)-pCOS:(Nth/2+1)+pCOS);

        % Solve the equation
        FTLobj = FTKerTrunc\FTLMTrunc;

        %Compute O_p/O_0 + offset. Offset does not affect standard deviation.
        % Note: O_p is in FTLobj(1) and O_0 is in FTLobj(pCOS+1) because of the truncation.  
        ratio(nSTD) = FTLobj(1)/FTLobj(pCOS+1) + offset;   
    end
    
    %Store results.
    DeltaPExpe(1,p) = pCOS;
    DeltaPExpe(2,p) = abs(std(ratio));
    
    if(p==p2Show)%Save to show an example.
       ratioExample = abs(ratio - offset); 
    end
end



%%% Compute mean and noise standard deviation (STD) of camera measurements %%%
%Note:"Nth/2"-th column of "dataForCamNoise" data is taken to compute noise STD
%  assuming this data's mean is roughly the same as average value of Lmeas.
LmeasAll = csvread(strcat(fNameLmeasAll,num2str(pCOSAll(numPCOS)),'.csv'));
dataForCamNoise = LmeasAll(:,Nth/2);
camNoiseSTD = std(dataForCamNoise)/sqrt(Nth);
disp('** Camera noise standard deviation');
disp(camNoiseSTD*sqrt(Nth));
%Compute mean of Lmeas
M0 = mean(mean(LmeasAll)); 
disp('** Signal mean M0');
disp(M0);



%%% Compute theoretical DeltaP from eq. (39) %%%

%p-range considered to compute gamma.
pMax = Nth/2-1;%If gammaApprox diverges, choose smaller value.
pStart = Nth/2-pMax+1;
pEnd = Nth/2+pMax+1;

%Load BRDF data
Spec = csvread(fNameSpec);
Diff = csvread(fNameDiff);

%Prepare Fourier transformed data.
Spp = fftshift(fft(Spec))/Nth;
Dp = fftshift(fft(Diff))/Nth;

%Compute approximated gamma from eq. (40).
denomiGamma = 1.;
for p=pStart:1:pEnd
    denomiGamma = denomiGamma + conj(Dp(p))*Dp(p)/Spp(p);
end
gammaApprox = abs(1./denomiGamma);
disp('** Gamma (the degree of specularity) from eq. (40).');
disp(gammaApprox);

%Compute theoretical DeltaP from eq. (39).
DeltaPTheo = zeros(2, pCOSAll(numPCOS));
idx = 0;
for p=Nth/2+1:1:Nth/2+pCOSAll(numPCOS)+1
    idx=idx+1;
    DeltaPTheo(1,idx) = idx-1;
    DeltaPTheo(2,idx) = abs(Spp(Nth/2+1)/Spp(p)*camNoiseSTD/(M0*gammaApprox));
end

%Compute effective SNR from eq. (42).
SNReff = M0*gammaApprox/camNoiseSTD;
disp('** SNReff/sqrt(Nth) from eq. (42).');
disp(SNReff/sqrt(Nth));



%%% Display results %%%

%Comparison of experimental result and theoretical result.
figure;
hold on;
plot(DeltaPExpe(1,:), DeltaPExpe(2,:), 'LineWidth', 2, 'Color', 'b', 'LineStyle', '-');
plot(DeltaPTheo(1,:), DeltaPTheo(2,:), 'LineWidth', 2, 'Color', 'r', 'LineStyle', '-');
hold off;
title('Relative noise error', 'FontSize', 16);
xlabel('Fourier terms p', 'FontSize', 14);
ylabel('Blue: experimental, Red: theoretical', 'FontSize', 14);
axis([1,pCOSAll(numPCOS), 0, max(max(DeltaPExpe(2,:)),max(DeltaPTheo(2,:)))]);
annotation('textbox', [0.15, 0.72, 0.1, 0.1], 'String', strcat('Gamma =', num2str(gammaApprox)), 'FontSize', 14);
annotation('textbox', [0.15, 0.8, 0.1, 0.1], 'String', strcat('SRNeff/sqrt(Nth) =', num2str((SNReff)/sqrt(Nth))), 'FontSize', 14);

%An example of 500 trials of measured signal Lmeas.
figure;
LmeasAllExample = csvread(strcat(fNameLmeasAll,num2str(pCOSAll(p2Show)),'.csv'));
surf(LmeasAllExample);
title(strcat('Example of measured signal Lmeas (', num2str(num2str(pCOSAll(p2Show))), '-period case)'), 'FontSize', 16);
xlabel('ThetaOut [deg]', 'FontSize', 14);
ylabel('Trials', 'FontSize', 14);
zlabel('Signal', 'FontSize', 14);
xticks([1 16 32 48 64]);
xticklabels({'29','37','45','53','61'});
axis([1,Nth, 1, numTrials, 0, max(max(LmeasAllExample))]);

%An example of 500 trials of the ratio O_p/O_0
figure;
plot(ratioExample, 'LineWidth', 1, 'Color', 'b', 'LineStyle', '-');
title(strcat('Example of experimental Op/O0 (', num2str(num2str(pCOSAll(p2Show))), '-period case)'), 'FontSize', 16');
xlabel('Trials', 'FontSize', 14);
