%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Light field reconstruction by the Wiener filter and the truncation
%
%Takahiro Sasaki, Jan 2020, sasa0027@umn.edu
% Scientific Computation Program, Electrical and Computer Engineering Dept.
% The University of Minnesota Twin Cities
%
%This code is a part of our JOSA A journal paper:
% Sasaki, T., & Leger, J., "Light field reconstruction
%  from scattered light using plenoptic data," JOSA A (2020)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%{
***How to use***

    1)Choose a scattering material by changing "sampleID".
        0: Brushed metal, 1: White paper #1, 2: White paper #2
    2)Execute.

***Data format and some details***

    -Angular domain range is [29deg,61deg] for both theta_in and theta_out.
     The number of angular data points Nth is 64, or angular increment is 0.5deg.

    -Lmeas.csv (Measured scattered signal):
        -Nth 1-D array.
        -A function of theta_out.
        -Lightly moving-averaged Lmeas is taken to compute the analytical
        expressions, because they might become inaccurate at higher frequencies.

    -BRDFCOS.csv (BRDF x COS(theta_in) of a scatter)
        -Nth x Nth 2-D array.
        -Row: Outgoing angle, or theta_out.
        -Column: Incoming angle, or theta_in.
        -This includes COS(theta_in) term.

    -Spec.csv (Specular component of BRDF):
        -This needs to be computed from BRDF w/o COS term.
        -See section 5.A for details.
        -Nth 1-D array.

    -Diff.csv (Diffuse component of BRDF):
        -This needs to be computed from BRDF w/o COS term.
        -The diffuse is expressed as (Diff)(x)(Diff) where (x) represents
         outer-product, and data for (Diff) is this file.
        -See section 5.A for details.
        -Nth 1-D array.

***Notes***

    -Wiener filter includes matrix truncation at "pHatMax X 1.5",
     because the filter may become inaccurate at higher frequencies.
     (pHatMax is the Fourier term predicted from eq. (46).)

    -The above BRDF-related files are slightly different
        from those for DeltaP.m tool.

    -The provided data and results may be slightly different
        from those reported in our JOSA A journal paper.

***For further details, see the following codes and values***
%}

clear;

%%%%%%%%%%%%%%% Choose scattering material %%%%%%%%%%%%%
%0: Brushed metal, 1: White paper #1, 2: White paper #2
sampleID = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%% Common Settings %%%
%Number of data points in theta (angular) direction
Nth = 64;

%Grid spacing in theta (angular) direction
dth = 0.5*pi/180.;%[rad]

%Tilt angle of the scattering plane
scattTilt = -45.*pi/180.;%[rad]

%%% Scatter dependent settings %%%
if(sampleID == 0)%Brushed metal scatter

    disp('*** Brushed metal scatter ***');
    
    %File name of Lmeas
    fNameLmeas = './BrushedMetal/LFRecon/LmeasMetal.csv';
    
    %Signal mean, i.e. average value of Lmeas.
    M0 = 5090;%Camera signal
    %Note: Only for brushed metal, M0 was computed from Lmeas(20:50)
    % to neglect low signal level part of Lmeas.
    
    %Measured camera-noise standard deviation (STD)
    % at the signal level of the above M0.
    camNoiseSTD =  110.5/sqrt(Nth);%Noise standard deviation/sqrt(Nth)
    %Note: No pixel averaging was applied in the measurement.

    %File name of BRDF x COS. (Note: This includes COS(theta_in) term.)
    fNameBRDFCOS = './BrushedMetal/LFRecon/BRDFCOSMetalA.csv';

    %File name of specular component of BRDF.
    fNameSpec = './BrushedMetal/LFRecon/SpecMetalA.csv';

    %File name of diffuse component of BRDF.
    %Note: This is separable expression. The diffuse is expressed as
    % (Diff)(X)(Diff) where (X) represents outer product, and (Diff) is this file:
    fNameDiff = './BrushedMetal/LFRecon/DiffMetalA.csv';
    
    %Experimentally determined truncation point to minimize MSE.
    pMSEMin = 8;
        
elseif(sampleID == 1)%White paper #1 scatter

    disp('*** White paper #1 scatter ***');
    
    %File name of Lmeas
    fNameLmeas = './WhitePaper1/LFRecon/LmeasWPaper1.csv';
    
    %Signal mean, i.e. average value of Lmeas.
    M0 = 5319;%Camera signal
    
    %Measured camera-noise standard deviation (STD)
    % at the signal level of the above M0.
    camNoiseSTD =  113.3/sqrt(Nth);%Noise standard deviation/sqrt(Nth)
    %Note: No pixel averaging was applied in the measurement.
    
    %File name of BRDF x COS. (Note: This includes COS(theta_in) term.)
    fNameBRDFCOS = './WhitePaper1/LFRecon/BRDFCOSWPaper1A.csv';

    %File name of specular component of BRDF.
    fNameSpec = './WhitePaper1/LFRecon/SpecWPaper1A.csv';

    %File name of diffuse component of BRDF. 
    %Note: This is separable expression. The diffuse is expressed as
    % (Diff)(X)(Diff) where (X) represents outer product, and (Diff) is this file:
    fNameDiff = './WhitePaper1/LFRecon/DiffWPaper1A.csv';
    
    %Experimentally determined truncation point to minimize MSE.
    pMSEMin = 6;

elseif(sampleID == 2)%White paper #2 scatter

    disp('*** White paper #2 scatter ***');

    %File name of Lmeas
    fNameLmeas = './WhitePaper2/LFRecon/LmeasWPaper2.csv';
    
    %Signal mean, i.e. average value of Lmeas.
    M0 = 1864;%Camera signal
    
    %Measured camera-noise standard deviation (STD)
    % at the signal level of the above M0.
    camNoiseSTD =  36.1/sqrt(Nth);%Noise standard deviation/sqrt(Nth)
    %Note: 3-pixel averaging was applied in the measurement.
    
    %File name of BRDF x COS. (Note: This includes COS(theta_in) term.)
    fNameBRDFCOS = './WhitePaper2/LFRecon/BRDFCOSWPaper2A.csv';

    %File name of specular component of BRDF.
    fNameSpec = './WhitePaper2/LFRecon/SpecWPaper2A.csv';

    %File name of diffuse component of BRDF. 
    %Note: This is separable expression. The diffuse is expressed as
    % (Diff)(X)(Diff) where (X) represents outer product, and (Diff) is this file:
    fNameDiff = './WhitePaper2/LFRecon/DiffWPaper2A.csv';
    
    %Experimentally determined truncation point to minimize MSE.
    pMSEMin = 3;
end
%%% End of Settings %%%



%%% Preparations for computation %%%

%Angular domain range
startDomain = abs(scattTilt)-dth*Nth/2 ;
endDomain = abs(scattTilt)+dth*Nth/2;
rangeDomain = endDomain - startDomain;

%Load measured BRDFxCOS data.
Ker = csvread(fNameBRDFCOS);

%Load Lmeas data and compute mean value
Lmeas = csvread(fNameLmeas);

%Load specular component of BRDF.
Spec = csvread(fNameSpec);

%Load diffuse component of BRDF.
Diff = csvread(fNameDiff);

%Lightly moving-averaged data is taken to compute the analytical expressions.
%(This is to remove inaccuracy of the analytical expressions at higher freq.)
LmeasWienerFirstDeriv = movmean(Lmeas, 5);

%Prepare Fourier transformed data.    
Spp = fftshift(fft(Spec))/Nth;
Dp = fftshift(fft(Diff))/Nth;
MpWienerFirstDeriv = fftshift(fft(LmeasWienerFirstDeriv))/Nth;

%p-range to be considered for the analytical expressions
pMax = Nth/2-1;%If alpha and/or gammaApprox diverge, choose smaller value.
pStart = Nth/2-pMax+1;
pEnd = Nth/2+pMax+1;



%%% Compute the constant alpha from eq. (34) %%%
denomiAlpha = 1.;
for p=pStart:1:pEnd
    denomiAlpha = denomiAlpha + abs(Dp(p))^2/Spp(p);
end
numeAlpha = 0.;
for p=pStart:1:pEnd
    numeAlpha = numeAlpha + conj(Dp(p))/Spp(p)*MpWienerFirstDeriv(p);
end
alpha = numeAlpha/denomiAlpha;

disp('** Constant alpha from eq. (34).');
disp(alpha);



%%% Compute gamma (the degree of specularity) from eq. (40) %%%
denomiGamma = 1.;
for p=pStart:1:pEnd
    denomiGamma = denomiGamma + conj(Dp(p))*Dp(p)/Spp(p);
end
gammaApprox = abs(1./denomiGamma);

disp('** Gamma (the degree of specularity) from eq. (40).');
disp(gammaApprox);

%Compute effective SNR from eq. (42).
SNReff = M0*gammaApprox/camNoiseSTD;
disp('** SNReff/sqrt(Nth) from eq. (42).');
disp(SNReff/sqrt(Nth));



%%% Compute the first derivative of MSE from eq. (46) %%%
%Note: pHatMax is the MSE minimizer explained in section 4.A and 5.D. 
firstDerivMSE = zeros(Nth/2, 1);

flug = 0;
maxPFDM = Nth/2;
idxP = 0;
for p=Nth/2+1:1:Nth
    idxP = idxP + 1;
    
    firstDerivMSE(idxP) = -abs((MpWienerFirstDeriv(p)-alpha*Dp(p))/Spp(p))^2 + camNoiseSTD^2/abs(Spp(p))^2;

    if(p==Nth/2+1)
        valFDM = abs(firstDerivMSE(idxP));
    end
   
    if(firstDerivMSE(idxP)>=0. && flug==0)
        flug = 1;
        pHatMax = idxP-1;
    end

    if(valFDM < firstDerivMSE(idxP))
        maxPFDM = idxP;
        break;
    end
end



%%% Compute Wiener filter %%%

%Compute beta0 from eq. (a2).
beta0 = 0;
for i=1:1:Nth
    theta = startDomain + dth*(i-1); 
    beta0 = beta0 + 1/cos(theta);
end
beta0 = beta0*dth/rangeDomain;

%Compute Wiener filter's regularizer, i.e. inverse of eq. (49).
wienerFilter = zeros(Nth,1);
for p=1:1:Nth
    wienerFilter(p) = abs(Spp(p))^2/abs(MpWienerFirstDeriv(p) - alpha*Dp(p))^2;
end
wienerFilter = wienerFilter*camNoiseSTD^2/(beta0^2);



%%% Solve equation by the two methods %%%
% method=1: Truncation regularization by pHatMax predicted from eq. (46)
% method=2: Truncation regularization by experimentally determined pMSEMin
% method=3: Wiener filter from eq. (48)
for method=1:1:3

    %%% Prepare matrix
    if(method==1)%For truncation regularization predicted from the MSE derivative.
        truncPoint = pHatMax;
        
    elseif(method==2)%For truncation regularization determined from experiments.
        truncPoint = pMSEMin+1;
        
    else%For Wiener filter
        truncPoint = ceil(pHatMax*1.5);
        if(truncPoint>Nth/2-1)
            truncPoint = Nth/2-1;
        end
    end

    if(method==3)%For Wiener filter
        KerOriginal = Ker; 
        Ker = transpose(KerOriginal)*KerOriginal*dth;
    end

    %2D-FT of Ker
    FTKer = fftshift(fft2(Ker))/(Nth*Nth);
    FTKer = rangeDomain*fliplr(FTKer);

    if(method==3)%For Wiener filter
        for j=1:1:Nth-1
            FTKer(j+1, j) = FTKer(j+1, j) + wienerFilter(j+1);
        end
    end

    %Prepare truncated matrix
    FTKerTrunc = FTKer((Nth/2+1)-truncPoint:(Nth/2+1)+truncPoint, (Nth/2)-truncPoint:(Nth/2)+truncPoint);

    %Condition number
    if(method==1)
        disp('** Condition number of matrix (Truncation at pHatMax)');
    elseif(method==2)
        disp('** Condition number of matrix (Truncation at pMSEMin)');        
    else
        disp('** Condition number of matrix (Wiener filter)');
    end
    disp(cond(FTKerTrunc));
    
    %%% Prepare  right-hand-side vector
    if(method==3)%For Wiener filter
        LmeasOriginal = Lmeas;
        Lmeas = transpose(KerOriginal)*LmeasOriginal;
    end

    %1D-FT of Lmeas
    FTLmeas = fftshift(fft(Lmeas))/Nth;

    %Prepare truncated data
    FTLmeasTrunc = FTLmeas(((Nth/2+1)-truncPoint):((Nth/2+1)+truncPoint));

    %%% Solve
    FTLobj = FTKerTrunc\FTLmeasTrunc;
    solution = ifft(FTLobj(2:truncPoint*2),Nth);
    solution = abs(solution)*Nth;
    
    if(method==1)%For truncation regularization by pHatMax
        LobjTruncPHatMax = solution;
    elseif(method==2)
        LobjTruncPMSEMinExpe = solution;
    else%For Wiener filter
        LobjWiener = solution;
    end
end



%%% Display results %%%

%Experimentally determined MSE minimum truncation result  
figure;
hold on
plot(LobjTruncPMSEMinExpe/max(LobjTruncPMSEMinExpe), 'LineWidth', 2, 'Color', 'b', 'LineStyle', '-');
plot(LmeasOriginal/max(LmeasOriginal), 'LineWidth', 2, 'Color', 'r', 'LineStyle', '--');
hold off
title('Experimentally determined MSE min. truncation result', 'FontSize', 16);
xlabel('Angle [deg]', 'FontSize', 14);
ylabel('Normalized intensity Red:Lmeas, Blue:Lobj', 'FontSize', 14);
xticks([1 16 32 48 64]);
xticklabels({'29','37','45','53','61'});
yticks([0]);
yticklabels({''});
axis([1,Nth, 0, 1]);
annotation('textbox', [0.15, 0.64, 0.1, 0.1], 'String', strcat('pMSEMin =', num2str(pMSEMin)), 'FontSize', 13);
annotation('textbox', [0.15, 0.72, 0.1, 0.1], 'String', strcat('Gamma =', num2str(gammaApprox)), 'FontSize', 13);
annotation('textbox', [0.15, 0.8, 0.1, 0.1], 'String', strcat('SNReff/sqrt(Nth) =', num2str(SNReff/sqrt(Nth))), 'FontSize', 13);

%Truncation result with theoretically predicted pHatMax
figure;
hold on
plot(LobjTruncPHatMax/max(LobjTruncPHatMax), 'LineWidth', 2, 'Color', 'b', 'LineStyle', '-');
plot(LmeasOriginal/max(LmeasOriginal), 'LineWidth', 2, 'Color', 'r', 'LineStyle', '--');
hold off
title('Truncation result with predicted pHatMax', 'FontSize', 18);
xlabel('Angle [deg]', 'FontSize', 14);
ylabel('Normalized intensity Red:Lmeas, Blue:Lobj', 'FontSize', 14);
xticks([1 16 32 48 64]);
xticklabels({'29','37','45','53','61'});
yticks([0]);
yticklabels({''});
axis([1,Nth, 0, 1]);
annotation('textbox', [0.15, 0.64, 0.1, 0.1], 'String', strcat('pHatMax =', num2str(pHatMax-1)), 'FontSize', 13);
annotation('textbox', [0.15, 0.72, 0.1, 0.1], 'String', strcat('Gamma =', num2str(gammaApprox)), 'FontSize', 13);
annotation('textbox', [0.15, 0.8, 0.1, 0.1], 'String', strcat('SNReff/sqrt(Nth) =', num2str(SNReff/sqrt(Nth))), 'FontSize', 13);

%Theoretical first derivative of MSE
figure;
hold on
pIdx = 0:1:(maxPFDM-1);
plot(pIdx, firstDerivMSE(1:maxPFDM)/valFDM, 'LineWidth', 2, 'Color', 'b', 'LineStyle', '-');
xAx = zeros(maxPFDM,1);
plot(pIdx, xAx, 'LineWidth', 1, 'Color', 'r', 'LineStyle', '-');
hold off
title('Theoretical first derivative of MSE', 'FontSize', 18);
xlabel('Fourier term p', 'FontSize', 14);
ylabel('Normalized', 'FontSize', 14);
axis([0,maxPFDM-1, -1, 1]);
annotation('textbox', [0.2, 0.8, 0.1, 0.1], 'String', strcat('pHatMax =', num2str(pHatMax-1)), 'FontSize', 14);

%Wiener filter result
figure;
hold on
plot(LobjWiener/max(LobjWiener), 'LineWidth', 2, 'Color', 'b', 'LineStyle', '-');
plot(LmeasOriginal/max(LmeasOriginal), 'LineWidth', 2, 'Color', 'r', 'LineStyle', '--');
hold off
title('Wiener filter result', 'FontSize', 18);
xlabel('Angle [deg]', 'FontSize', 14);
ylabel('Normalized intensity Red:Lmeas, Blue:Lobj', 'FontSize', 14);
xticks([1 16 32 48 64]);
xticklabels({'29','37','45','53','61'});
yticks([0]);
yticklabels({''});
axis([1,Nth, 0, 1]);
annotation('textbox', [0.15, 0.72, 0.1, 0.1], 'String', strcat('Gamma =', num2str(gammaApprox)), 'FontSize', 13);
annotation('textbox', [0.15, 0.8, 0.1, 0.1], 'String', strcat('SNReff/sqrt(Nth) =', num2str(SNReff/sqrt(Nth))), 'FontSize', 13);

%Wiener filter matrix
figure;
hold on
surf(log(abs(FTKerTrunc)));
hold off
title('Wiener filter matrix', 'FontSize', 18);
xlabel('s', 'FontSize', 14);
ylabel('p', 'FontSize', 14);
zlabel('Logarithm of each element', 'FontSize', 14);
view(-25,45);
zticks([0]);
axis([1, truncPoint*2+1, 1, truncPoint*2+1, min(min(log(abs(FTKerTrunc)))), max(max(log(abs(FTKerTrunc))))]);
